# cython: language_level=3str, binding=True
# Copyright (c) 2012-20 G. Peter Lepage.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version (see <http://www.gnu.org/licenses/>).
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

cimport numpy

from numpy cimport npy_intp as INTP_TYPE
# index type for numpy (signed) -- same as numpy.intp_t and Py_ssize_t

cdef packed struct svec_element:
    double v
    INTP_TYPE i

cdef class svec:
    cdef svec_element * v
    cdef readonly int size
    cpdef numpy.ndarray[numpy.float_t, ndim=1] toarray(svec,INTP_TYPE msize=?)
    cpdef numpy.ndarray[INTP_TYPE, ndim=1] indices(svec)
    cpdef numpy.ndarray[numpy.float_t, ndim=1] values(svec)
    cpdef _assign(self, const numpy.float_t[:], const INTP_TYPE[:])
    cpdef double dot(svec, svec)
    cpdef svec clone(svec)
    cpdef svec add(svec,svec,double a=*,double b=*)
    cpdef svec mul(svec self,double a)
    cpdef numpy.ndarray[numpy.float_t, ndim=1] masked_vec(svec, smask, out=*)

cdef class smask:
    cdef readonly numpy.int8_t[::1] mask
    cdef readonly INTP_TYPE[::1] map 
    cdef readonly INTP_TYPE starti, stopi, len
 
cdef class smat:
    cdef object[::1] row
    # cdef svec[::1] row
    cdef INTP_TYPE[::1] block
    cdef INTP_TYPE nrow, nrow_max, next_block
    cpdef _add_memory(smat self)
    cpdef INTP_TYPE blockid(smat self, INTP_TYPE i)
    cpdef numpy.ndarray[INTP_TYPE, ndim=1] append_diag(self, const numpy.float_t[:])
    cpdef numpy.ndarray[INTP_TYPE, ndim=1] append_diag_m(self, const numpy.float_t[:,:])
    cpdef add_offdiag_m(self, const numpy.npy_intp[:], const numpy.npy_intp[:], const numpy.float_t[:, :])
    cpdef svec dot(self,svec)
    cpdef svec masked_dot(self, svec vv, const numpy.int8_t[:] imask)
    cpdef double expval(self,svec)
    cpdef numpy.ndarray[numpy.float_t, ndim=2] toarray(self)
    cpdef numpy.ndarray[numpy.float_t, ndim=2] masked_mat(smat self, smask mask, out=*)



"""
Module :mod:`gvar.powerseries` provides tools for manipulating power series
approximations of functions. A function's power series is specified by the
coefficients in its Taylor expansion with respect to an independent variable,
say ``x``::

    f(x) = f(0) + f'(0) * x + (f''(0)/2) * x**2 + (f'''(0)/6) * x**3 + ...
         = f0 + f1 * x + f2 * x**2 + f3 * x**3 + ...

In practice a power series is different from a polynomial because power
series, while infinite order in principle, are truncated at some finite
order in numerical applications. The ``order`` of a power series is the
highest power of ``x`` that is retained in the approximation; coefficients
for still higher-order terms are assumed to be unknown (as opposed to zero).

Taylor's theorem can be used to generate power series for functions of
power series::

    g(f(x)) = g(f0) + g'(f0) * (f(x)-f0) + (g''(f0)/2) * (f(x)-f0)**2 + ...
            = g0 + g1 * x + g2 * x**2 + ...

This allows us to define a full calculus for power series, where arithmetic
expressions and (sufficiently differentiable) functions of power series
return new power series.

Using power series
-----------------------
Class :class:`PowerSeries` provides a numerical implementation of the power series
calculus. ``PowerSeries([f0,f1,f2,f3...])`` is a numerical representation of
a power series with coefficients ``f0, f1, f2, f3...`` (as in ``f(x)``
above). Thus, for example, we can define a 4th-order power series
approximation ``f`` to ``exp(x)=1+x+x**2/2+...`` using

    >>> import gvar as gv 
    >>> import gvar.powerseries as ps
    >>> f = ps.PowerSeries([1., 1., 1/2., 1/6., 1/24.])
    >>> print(f)            # print the coefficients
    [ 1.          1.          0.5         0.16666667  0.04166667]

Arithmetic expressions involving instances of class :class:`PowerSeries` are
themselves :class:`PowerSeries` as in, for example,

    >>> print(1/f)              # power series for exp(-x)
    [ 1.         -1.          0.5        -0.16666667  0.04166667]
    >>> print(gv.log(f))        # power series for x
    [ 0.  1.  0. -0.  0.]
    >>> print(f / f)            # power series for 1
    [ 1.  0.  0.  0.  0.]

The standard arithmetic operators (``+,-,*,/,=,**``) are supported, as are
the usual elementary functions (``exp, log, sin, cos, tan ...``). Different
:class:`PowerSeries` can be combined arithmetically to create new
:class:`PowerSeries`; the order of the result is that of the operand with the
lowest order.

:class:`PowerSeries` can be differentiated and integrated::

    >>> print(f.deriv())    # derivative of exp(x)
    [ 1.          1.          0.5         0.16666667]
    >>> print(f.integ())    # integral of exp(x) (from x=0)
    [ 0.          1.          0.5         0.16666667  0.04166667  0.00833333]

Each :class:`PowerSeries` represents a function. The :class:`PowerSeries` for
a function of a function is easily obtained. For example, assume ``f``
represents function ``f(x)=exp(x)``, as above, and ``g``
represents ``g(x)=log(1+x)``::

    >>> g = ps.PowerSeries([0, 1, -1/2., 1/3., -1/4.])

Then ``f(g)`` gives the :class:`PowerSeries` for ``exp(log(1+x)) = 1 + x``::

    >>> print(f(g))
    [  1.0000e+00   1.0000e+00   0.0000e+00  -2.7755e-17 -7.6327e-17]

Individual coefficients from the powerseries can be accessed using
array-element notation: for example,

    >>> print(f[0], f[1], f[2], f[3])
    1.0 1.0 0.5 0.166666666667
    >>> f[0] = f[0] - 1.
    >>> print(f)            # f is now the power series for exp(x)-1
    [ 0.          1.          0.5         0.16666667  0.04166667]

Finally, a power series can be evaluated for a particular
numerical value of x:

    >>> x = 0.01
    >>> print(f(x))             # should be exp(0.01)-1 approximately
    0.0100501670833
    >>> print(gv.exp(x)-1)      # verify that it is
    0.0100501670842

The independent variable ``x`` could be of any arithmetic type (it need not
be a ``float``).

Multivariate power series
-------------------------
The coefficients in a :class:`PowerSeries` object can themselves by :class:`PowerSeries` 
objects. The is used to represent multivariate power series such as::

    f(x,y) = f00 + f10 * x + f01 * y + f20 * x**2 + f11 * x*y + f02 * y**2 + ...

One way to construct a :class:`PowerSeries` object ``f`` representing this series, through
``order=2``, is from an array ``c`` containing the coefficients::

    c = [[f00, f01, f02], [f10, f11, 0], [f20, 0, 0]]
    f = ps.multiseries(c, order=2)

Here entries for ``c[1,2]``, ``c[2,1]``,  and ``c[2,2]`` are ignored because they 
correspond to ``order=3`` or higher.
The individual coefficients ``c[i,j]`` are accessed using ``f[i,j]``, and the 
function is evaluate at point ``(x,y)`` using ``f(x,y)``. Similarly the first-order 
partial derivative with respect to ``x`` and ``y``, for example, is given by 
:class:`PowerSeries` object
``f.deriv(1,1)``, while first-order integrals with respect to ``x`` and ``y`` 
are given by ``f.integ(1,1)``.

Taylor expansions of Python functions
-------------------------------------
:class:`PowerSeries` can be used to compute Taylor series for more-or-less
arbitrary pure-Python functions provided the functions are
sufficiently differentiable. To compute the ``N``-th order
expansion of a Python function ``g(x)``, first create a ``N``-th order
:class:`PowerSeries` variable that represents the expansion parameter: say,
``x = PowerSeries([0.,1.],order=N)``. The Taylor series for function ``g``
is then given by ``g_taylor = g(x)`` which is a :class:`PowerSeries` instance.
For example, consider::

    >>> def g(x):              # an example of a Python function
    ...     return 0.5/sqrt(1+x) + 0.5/sqrt(1-x)
    ...
    >>> x = ps.PowerSeries([0.,1.],order=5)    # Taylor series for x
    >>> print(x)
    [ 0.  1.  0.  0.  0.  0.]
    >>> g_taylor = g(x)        # Taylor series for g(x) about x=0
    >>> print(g_taylor)
    [ 1.         0.         0.375      0.         0.2734375  0.       ]
    >>> exp_taylor = gv.exp(x) # Taylor series for exp(x) about x=0
    >>> print(exp_taylor)
    [ 1.          1.          0.5         0.16666667  0.04166667  0.00833333]

This generalizes easily to multivariate expansions. For example, 
one can calculate the Taylor expansion coefficients for ``exp(x+y)`` using::

    >>> x,y = ps.multivar(dim=2, order=3)
    >>> exp_taylor = gv.exp(x + y)
    >>> print(exp_taylor)
    [[1.         1.         0.5        0.16666667], [1.  1.  0.5], [0.5 0.5], [0.16666667]]
    >>> print(exp_taylor(.1,.2), gv.exp(.1 + .2))
    1.3495000000000001 1.3498588075760032

Here function :func:`multivar` creates :class:`PowerSeries` objects corresponding to the 
expansion variables through a given order.
"""
# Created by G. Peter Lepage on 2009-12-14.
# Copyright (c) 2009-2020 G. Peter Lepage.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version (see <http://www.gnu.org/licenses/>).
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy
# import math
from numpy import exp, log, sqrt, sin, cos, tan, arcsin, arccos, arctan
from numpy import sinh, cosh, tanh, arcsinh, arccosh, arctanh
from scipy.special import eval_hermite as _scipy_eval_hermite

class PowerSeries(object):
    """
    Power series representation of a function.

    The power series created by ``PowerSeries(c)`` corresponds to::

        c[0] + c[1]*x + c[2]*x**2 + ... .

    The order of the power series is normally determined by the length of
    the input list ``c``. This can be overridden by specifying the order of
    the power series using the ``order`` parameter. The list of ``c[i]``\s
    is then padded with zeros if ``c`` is too short, or truncated if it
    is too long. Omitting ``c`` altogether results in a power series
    all of whose coefficients are zero. Individual series
    coefficients are accessed using array/list notation: for example,
    the 3rd-order coefficient of ``PowerSeries p`` is ``p[3]``. The
    order of ``p`` is ``p.order``. :class:`PowerSeries` should work
    for coefficients of any data type that supports ordinary arithmetic.

    Arithmetic expressions of :class:`PowerSeries` variables yield new
    :class:`PowerSeries` results that represent the power series
    expansion of the expression. Expressions can include the standard
    mathematical functions (``log, exp, sqrt, sin, cos, tan...``).
    :class:`PowerSeries` can also be differentiated (``p.deriv()``)
    and integrated (``p.integ()``).

    Args:
        c (array): Power series coefficients.
        order (int or None): Highest power in power series. If ``None``, 
            the order is inferred from the array of coefficients. 
            If array ``c`` is too small for the specified ``order``, 
            the array is padded with zeros at the end.
    """
    def __init__(self, c=None, order=None):
        if isinstance(c, PowerSeries):
            c = numpy.array(c.c)
        if order is None:
            if c is None:
                raise ValueError('must specify c or order')
            elif len(c) == 0:
                raise ValueError("empty coefficient array: c = "+str(c))
            order = len(c) - 1
        if order < 0:
            raise ValueError(
                "order cannot be less than zero: order = " +
                str(order)
                )
        if c is None or len(c) == 0:
            self.c = numpy.zeros(order + 1, object)
        elif order < len(c):
            self.c = numpy.array(c[:order + 1])
        elif order >= len(c):
            c = numpy.asarray(c)
            shape = (order + 1,) + c.shape[1:]
            self.c = numpy.zeros(shape, c.dtype)
            self.c[:len(c)] = c
            self.c[len(c):] *= c[0]
        # check for PowerSeries, fix orders
        for i in range(len(self.c)):
            if isinstance(self.c[i], PowerSeries):
                self.c[i] = PowerSeries(self.c[i].c, order=order - i)

    def _getorder(self):
        return len(self.c)-1
    order = property(_getorder,doc="Highest power in power series.")

    def _getcoeff(self):
        return numpy.array(self.c)
    coeff = property(_getcoeff,
                doc="Copy of power series coefficients (numpy.array).")

    def __getitem__(self, i):
        """ Return C{i}th coefficient of power series. """
        if numpy.size(i) > 1:
            return self.c[i[0]][i[1:]]
        else:
            return self.c[i if numpy.shape(i) == () else i[0]]

    def __setitem__(self, i, val):
        """ Set C{i}th coefficient of power series equal to C{val}. """
        if numpy.size(i) > 1:
            self.c[i[0]][i[1:]] = val
        else:
            self.c[i if numpy.shape(i) == () else i[0]] = val

    def __iter__(self):
        """ Iterate over coefficients of power series C{self}. """
        return iter(self.c)

    def __add__(self,x):
        try:
            order = min(self.order,x.order)
            ans = self.c[:order+1]+x.c[:order+1]
        except AttributeError:
            ans = list(self.c)
            ans[0] = ans[0] + x
        return PowerSeries(ans)

    def __radd__(self,x):
        return self+x

    def __sub__(self,x):
        try:
            order = min(self.order,x.order)
            ans = self.c[:order+1]-x.c[:order+1]
        except AttributeError:
            ans = list(self.c)
            ans[0] = ans[0] - x
        return PowerSeries(ans)

    def __neg__(self):
        return PowerSeries(-self.c)

    def __rsub__(self,x):
        return -self+x

    def __mul__(self,x):
        if isinstance(x,PowerSeries):
            order = min(self.order,x.order)
            ans = numpy.convolve(self.c,x.c,mode='full')[:order+1]
        else:
            ans = x*self.c
        return PowerSeries(ans)

    def __rmul__(self,x):
        return self*x

    # truediv and div are the same --- 1st is for python3, 2nd for python2
    def __truediv__(self,x):
        try:
            if x.c[0] == 0:     # needed for numpy scalars
                raise ZeroDivisionError
            order = min(self.order,x.order)
            ans = 0.0*self.c[:order+1]
            for n in range(order+1):
                tot = self.c[n]
                for i in range(n):
                    tot = tot - ans[i]*x.c[n-i]
                ans[n] = tot/x.c[0]
        except ZeroDivisionError:
            if self.c[0]==x.c[0] and len(self.c)>1 and len(x.c)>1:
                # strip off matching overall factor of x from denom and num
                return PowerSeries(self.c[1:])/PowerSeries(x.c[1:])
            else:
                raise ZeroDivisionError("leading coefficients in denominator equal zero")
        except AttributeError:
            ans = (1./x)*self.c
        return PowerSeries(ans)

    def __div__(self,x):
        try:
            if numpy.any(x.c[0] == 0):     # needed for numpy scalars
                raise ZeroDivisionError
            order = min(self.order,x.order)
            ans = 0.0*self.c[:order+1]
            for n in range(order+1):
                tot = self.c[n]
                for i in range(n):
                    tot = tot - ans[i]*x.c[n-i]
                ans[n] = tot/x.c[0]
        except ZeroDivisionError:
            if self.c[0]==x.c[0] and len(self.c)>1 and len(x.c)>1:
                # strip off matching overall factor of x from denom and num
                return PowerSeries(self.c[1:])/PowerSeries(x.c[1:])
            else:
                raise ZeroDivisionError("leading coefficients in denominator equal zero")
        except AttributeError:
            ans = (1./x)*self.c
        return PowerSeries(ans)

    def __rtruediv__(self,x):
        num = self.c*0
        num[0] = x
        return PowerSeries(num)/self

    def __rdiv__(self,x):
        num = self.c*0
        num[0] = x
        return PowerSeries(num)/self

    def __pow__(self,alpha):
        # use Taylor expn around x=c[0] to get answer
        # unless alpha is an integer (then multiply out)
        if numpy.issubdtype(type(alpha), numpy.int_):
            if alpha == 0:
                return 1.
            elif alpha < 0:
                alpha = - alpha
                xn = 1. / self
            else:
                xn = self
            ans = 1
            while alpha > 0:
                if alpha % 2 == 1:
                    ans *= xn
                xn *= xn
                alpha = (alpha - (alpha % 2)) / 2
            return ans
        fac = 1.
        alj = alpha
        xj = 1.
        ans = 0
        x = self-self.c[0]
        for j in range(self.order+1):
            ans = ans + fac*self.c[0]**alj*xj
            xj = xj*x
            fac = fac*alj/(j+1.)
            alj = alj-1
            # if fac==0:
            #     break
        return ans

    def __rpow__(self,x):
        return exp(self*log(x))

    def __str__(self):
        if isinstance(self.c[0], PowerSeries):
            return str([str(ci) for ci in self.c]).replace('"','').replace("'",'') # .replace(',','')
        else:
            return str(self.c)

    def __repr__(self):
        return "PowerSeries(%s)" % str(self.c.tolist())

    def __call__(self, *x):
        if len(x) > 1:
            return self(x[0])(*x[1:])
        elif len(x) == 1:
            x = x[0]
        ans = 0.0
        xn = 1.
        if isinstance(self.c[0], PowerSeries):
            # multivariate power series
            order = self.c[0].order
            for ci in self.c:
                ans += PowerSeries(ci, order=order) * xn
                xn *= x        
        else:
            for ci in self.c:
                ans += ci * xn
                xn *= x
        return ans

    def sqrt(self):
        return self**0.5

    def erf(self):
        from gvar import erf as _gvar_erf 
        jmax = self.order + 1
        jarray = numpy.arange(jmax + 1)
        deriv = (
            _scipy_eval_hermite(jarray, self.c[0])
            * (-1) ** jarray
            ) * (2 * numpy.exp(-self.c[0] ** 2) / numpy.sqrt(numpy.pi))
        x = self - self.c[0]
        fac = 1.
        ans = _gvar_erf(self.c[0])
        for j in range(1, jmax + 1):
            fac *= x / j
            ans = ans + fac * deriv[j - 1]
        return ans

    def sin(self):
        # use Taylor series about x=self.c[0]
        sc = sin(self.c[0])
        cc = cos(self.c[0])
        deriv = [sc,cc,-sc,-cc]
        fac = 1.
        xj = 1.
        ans = 0
        x = self-self.c[0]
        for j in range(self.order+1):
            ans = ans + fac*deriv[j%4]*xj
            fac = fac/(j+1.)
            xj = xj*x
        return ans

    def cos(self):
        # use Taylor series about x=self.c[0]
        sc = sin(self.c[0])
        cc = cos(self.c[0])
        deriv = [cc,-sc,-cc,sc]
        fac = 1.
        xj = 1.
        ans = 0
        x = self-self.c[0]
        for j in range(self.order+1):
            ans = ans + fac*deriv[j%4]*xj
            fac = fac/(j+1.)
            xj = xj*x
        return ans

    def tan(self):
        return sin(self)/cos(self)

    def arcsin(self):
        # use Taylors theorm to expand around x=self.c[0]
        # compute higher derivs from powerseries for 1st deriv
        x = PowerSeries([0.,1.],order=self.order-1)
        dasin = 1/sqrt(1-(self.c[0]+x)**2)
        # if True in numpy.isnan(dasin.c): # in dasin.c:
        #     raise ValueError("bad expansion point: "+str(self.c[0]))

        ans = arcsin(self.c[0])
        x = self-self.c[0]
        xn = x
        for n in range(1,self.order+1):
            ans = ans + (dasin.c[n-1]/float(n))*xn
            xn = xn*x
        return ans

    def arccos(self):
        # use Taylors theorm to expand around x=self.c[0]
        # compute higher derivs from powerseries for 1st deriv
        x = PowerSeries([0.,1.],order=self.order-1)
        dacos = -1/sqrt(1-(self.c[0]+x)**2)
        # if True in numpy.isnan(dacos.c): # in dasin.c:
        #     raise ValueError("bad expansion point: "+str(self.c[0]))

        ans = arccos(self.c[0])
        x = self-self.c[0]
        xn = x
        for n in range(1,self.order+1):
            ans = ans + (dacos.c[n-1]/float(n))*xn
            xn = xn*x
        return ans

    def arctan(self):
        # use Taylors theorm to expand around x=self.c[0]
        # compute higher derivs from powerseries for 1st deriv
        x = PowerSeries([0.,1.],order=self.order-1)
        datan = 1/(1+(self.c[0]+x)**2)

        ans = arctan(self.c[0])
        x = self-self.c[0]
        xn = x
        for n in range(1,self.order+1):
            ans = ans + (datan.c[n-1]/float(n))*xn
            xn = xn*x
        return ans

    def sinh(self):
        # use Taylor series about x=self.c[0]
        sc = sinh(self.c[0])
        cc = cosh(self.c[0])
        deriv = [sc,cc]
        fac = 1.
        xj = 1.
        ans = 0
        x = self-self.c[0]
        for j in range(self.order+1):
            ans = ans + fac*deriv[j%2]*xj
            fac = fac/(j+1.)
            xj = xj*x
        return ans

    def cosh(self):
        # use Taylor series about x=self.c[0]
        sc = sinh(self.c[0])
        cc = cosh(self.c[0])
        deriv = [cc,sc]
        fac = 1.
        xj = 1.
        ans = 0
        x = self-self.c[0]
        for j in range(self.order+1):
            ans = ans + fac*deriv[j%2]*xj
            fac = fac/(j+1.)
            xj = xj*x
        return ans

    def tanh(self):
        return sinh(self)/cosh(self)

    def arcsinh(self):
        return log(self+sqrt(self**2+1))

    def arccosh(self):
        return log(self+sqrt(self**2-1))

    def arctanh(self):
        return log((1+self)/(1-self))/2.

    def exp(self):
        f = exp(self.c[0])
        ans = 1.
        x = self - self.c[0]
        for n in range(self.order, 0, -1):
            ans = 1. + x * ans/ float(n)
        return f * ans

    def log(self):
        ans = log(self.c[0])
        x = -(self-self.c[0])/self.c[0]
        xn = 1
        for n in range(1,self.order+1):
            xn = xn*x
            ans = ans - xn/float(n)
        # ans[0] = ans[0] + log(self.c[0])
        return ans

    def deriv(self, *n):
        """ Compute *n*-th derivative (or partial derivative) of ``self``.

        Args:
            n (array): Number of derivatives in each direction.  Default is ``n=[1]``.

        Returns:
            :class:`PowerSeries` object representing the 
            *n*-th derivative or partial derivative of ``self``.
        """
        if len(n) == 0:
            n = [1]
        if len(n) > 1:
            ps = self.deriv(n[0])
            if ps == 0:
                return 0
            c = numpy.array(ps.c)
            for i in range(len(c)):
                if isinstance(c[i], PowerSeries):
                    c[i] = c[i].deriv(*n[1:])
                else:
                    c[i] = 0.0
            order = ps.order - sum(n[1:])
            if order >= 0:
                return PowerSeries(c, order=order)
            else:
                return 0.
        elif len(n) == 1:
            n = n[0]
        if n==1:
            if self.order > 0:
                return PowerSeries(self.c[1:] * range(1, len(self.c)))
            else:
                return 0.
        elif n>1:
            ans = self.deriv(1)
            return ans.deriv(n-1) if isinstance(ans, PowerSeries) else 0.
        elif n==0:
            return self

    def integ(self, *n, **kargs): # should be (self, *n, x0=1) but not with Py2
        """ Compute *n*-th indefinite integral of ``self``.

        If *x0* is specified, then the definite integral,
        integrating from point *x0*, is returned.

        Args:
            n (array): Number of integrations in each direction. Default is ``n=[1]``.
            x0 (array or float): Starting point for definite integral 
                in each direction (default is 0).
        Returns:
            :class:`PowerSeries` object representing the *n*-th integral of ``self``.
        """
        x0 = kargs.get('x0', 0.)
        if len(n) > 1:
            if numpy.shape(x0) == ():
                x0 = numpy.zeros(len(n), dtype=float) + x0 
            ps = self.integ(n[0], x0=x0[0])
            c = numpy.array(ps.c) 
            for i in range(len(c)):
                c[i] = c[i].integ(*n[1:], x0=x0[1:])
            order = ps.order + sum(n[1:])
            return PowerSeries(c, order=order)
        elif len(n) == 0:
            n = 1
        else:
            n = n[0]
        if numpy.shape(x0) != ():
            x0 = x0[0]
        if n==1:
            if self.order<0:
                ans = PowerSeries([0])
            else:
                ans = PowerSeries([0.*self.c[0]] +
                    [x/(i+1.) for i,x in enumerate(self.c)])
                return PowerSeries([ans.c[0]-ans(x0)]+list(ans.c[1:]))
        elif n>1:
            return self.integ().integ(n-1, x0=x0)
        elif n==0:
            return self


def multiseries(c, order=None):
    """ Create multivariate power series from coefficients in array ``c``.

    Args:
        c (array): :mod:`numpy`-like array containing the power series 
            coefficients. In ``d`` dimensions, ``c[i1,i2,...,id]`` is the 
            coefficient of ``x1**i1 * x2**i2 * ... * xd**id``.
        order (int or None): Highest power in power series, where the power 
            associated with term ``x1**i1 * x2**i2 * ... * xd**id`` 
            is the sum of the exponents: ``i1 + i2 + ... + id``. If ``None``,
            the order is inferred from the array of coefficients.
    
    Returns:
        :class:`PowerSeries` object representing the multivariate power series. 
    """
    c = numpy.asarray(c)
    if c.ndim <= 0:
        return None
    if c.size == 0:
        c = numpy.zeros(c.ndim * (1,), dtype=c.dtype)
    if order is None:
        order = sum([n - 1 for n in c.shape])
    ans = []
    for i, ci in enumerate(c[:order + 1]):
        if ci.shape == ():
            ans.append(ci)
        else:
            ans.append(multiseries(ci, order=order - i))
    ans = PowerSeries(ans, order=order)
    ans.ndim = c.ndim
    return ans
    
def multivar(dim, order):
    """ Create :class:`PowerSeries` objects representing the expansion variables.

    Args:
        dim (int): The dimensionality of the multivariate space.
        order (int): Highest power in the power series,  where the power 
            associated with term ``x1**i1 * x2**i2 * ... * xd**id`` 
            is the sum of the exponents: ``i1 + i2 + ... + id``.

    Returns:
        An array of ``dim`` :class:`PowerSeries` objects corresponding to 
        each of the expansion variables in a ``dim``-dimensional multivariate
        power series.
    """
    if dim <= 0 or order < 0:
        return []
    shape = dim * (order + 1,)
    ans = numpy.empty(dim, dtype=object)
    idx = numpy.zeros(dim, dtype=int)
    zeros = numpy.zeros(shape, dtype=float)
    for i in range(dim):
        ans[i] = multiseries(zeros, order=order)
        idx[i] = 1
        ans[i][idx] = 1.
        idx[i] = 0
    return ans



# cython: language_level=3str, binding=True
# Created by G. Peter Lepage (Cornell University) on 2012-05-31.
# Copyright (c) 2012-22 G. Peter Lepage.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version (see <http://www.gnu.org/licenses/>).
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import collections
import re
import numpy
import copy
import pickle
import json
import gvar as _gvar
import sys 

try:
    from collections.abc import MutableMapping as collections_MMapping
except ImportError:
    from collections import MutableMapping as collections_MMapping

if sys.version_info.major == 2:
    ORDEREDDICT = collections.OrderedDict
else:
    if sys.version_info.major == 3 and sys.version_info.minor < 6:
        ORDEREDDICT = collections.OrderedDict
    else:
        ORDEREDDICT = dict

BUFFERDICTDATA = collections.namedtuple('BUFFERDICTDATA',['slice','shape'])
""" Data type for BufferDict._data[k]. Note shape==() implies a scalar. """

class BufferDict(collections_MMapping):
    """ Ordered dictionary whose values are packed into a 1-d buffer (:mod:`numpy` array).

    |BufferDict|\s can be created in the usual way dictionaries are created::

        >>> b = BufferDict()
        >>> b['a'] = 1
        >>> b['b'] = [2.7, 3.2]
        >>> print(b)
        {'a': 1.0, 'b': array([2.7, 3.2])}
        >>> print(b.buf)
        [1. 2.7 3.2]
        >>> b = BufferDict(a=1, b=[2.7, 3.2])
        >>> b = BufferDict([('a',1.), ('b',[2.7, 3.2])])

    They can also be created from other dictionaries or |BufferDict|\s::

        >>> c = BufferDict(b)
        >>> print(c)
        {'a': 1.0, 'b': 2.7}
        >>> c = BufferDict(b, keys=['b'])
        >>> print(c)
        {'b': 2.7}

    The ``keys`` keyword restricts the copy of a dictionary to entries whose
    keys are in ``keys``.

    The values in a |BufferDict| are scalars or arrays of a scalar type
    (|GVar|, ``float``, ``int``, etc.). The data type of the buffer is 
    normally inferred (dynamically) from the data itself::

        >>> b = BufferDict(a=1)
        >>> print(b, b.dtype)
        {'a': 1} int64
        >>> b['b'] = 2.7
        >>> print(b, b.dtype)
        {'a': 1.0, 'b': 2.7} float64     
    
    The data type of the |BufferDict|'s buffer can be specified
    when creating the |BufferDict| from another dictionary or list
    by using keyword ``dtype``::

        >>> b = BufferDict(dict(a=1.2), dtype=int)  
        >>> print(b, b.dtype)
        {'a': 1} int64        
        >>> b['b'] = 2.7
        >>> print(b, b.dtype)
        {'a': 1, 'b': 2} int64  
  
    Note in this example that the data type is *not* changed by subsequent
    additions to the |BufferDict| when the ``dtype`` keyword is specified.
    To create an empty |BufferDict| with a specified type use, for example,
    ``BufferDict({}, dtype=int)``. Any data type supported by :mod:`numpy` 
    arrays can be specified.

    Some simple arithemetic is allowed between two |BufferDict|\s, say,
    ``g1`` and ``g2`` provided they have the same keys and array shapes.
    So, for example::

        >>> a = BufferDict(a=1, b=[2, 3])
        >>> b = BufferDict(b=[20., 30.], a=10.)
        >>> print(a + b)
            {'a': 11.0, 'b': array([22., 33.])}

    Subtraction is also defined, as are multiplication and division
    by scalars. The corresponding ``+=``, ``-=``, ``*=``, ``/=`` operators
    are supported, as are unary ``+`` and ``-``.

    Finally a |BufferDict| can be cloned from another one using
    a different buffer (containing different values)::

        >>> b = BufferDict(a=1., b=2.)
        >>> c = BufferDict(b, buf=[10, 20])
        >>> print(c)
        {'a': 10, 'b': 20}

    The  new buffer itself (i.e., not a copy) is used in the 
    new |BufferDict| if the buffer is a :mod:`numpy` array:

        >>> import numpy as np
        >>> b = BufferDict(a=1., b=2.)
        >>> nbuf = np.array([10, 20])
        >>> c = BufferDict(b, buf=nbuf)
        >>> c['a'] += 1
        >>> print(c)
        {'a': 11, 'b': 20}
        >>> print(nbuf)
        [11 20]

    """

    extension_pattern = re.compile('^([^()]+)\((.+)\)$')
    invfcn = {}

    def __init__(self, *args, **kargs):
        self._extension = {}
        self._ver_l = BufferDict._ver_g
        self.shape = None
        self._dtype = None     # enforced dtype (default is None)
        if len(args)==0:
            # kargs are dictionary entries
            self._odict = ORDEREDDICT()
            self._buf = numpy.array([])
            for k in kargs:
                self[k] = kargs[k]
            return
        if len(args) > 2:
            raise RuntimeError('wrong number of arguments')
        self._dtype = kargs.get('dtype', None)
        keys = list(kargs['keys']) if 'keys' in kargs else None
        if 'buf' in kargs:
            buf = kargs['buf']
            if len(args) > 1:
                raise RuntimeError('too many arguments: buffer specified more than once')
        elif len(args) == 2:
            buf = args[1]
        else:
            buf = None
        if isinstance(args[0], BufferDict) and keys is None:
            # make copy of BufferDict args[0], possibly with new buffer
            # copy keys, slices and shapes (fast build)
            bd = args[0]
            self._odict = ORDEREDDICT(bd._odict)
            # copy buffer or use new one
            if buf is None:
                self._buf = numpy.array(bd._buf, dtype=self._dtype)
            elif numpy.shape(buf) == bd._buf.shape:
                self._buf = numpy.asarray(buf, dtype=self._dtype)
            else:
                raise ValueError("buf is wrong shape --- %s not %s"
                                    % (numpy.shape(buf), bd._buf.shape))
            return
        if buf is not None:
            raise RuntimeError("can't specify buffer unless first argument is a BufferDict")
        data = args[0].items() if hasattr(args[0], 'keys') else args[0]
        self._odict = ORDEREDDICT()
        self._buf = numpy.array([], dtype=self._dtype) 
        if keys is None:
            for k, v in data:
                self[k] = v
        else:
            for k, v in data:
                if k in keys:
                    self[k] = v 

    def __copy__(self):
        return BufferDict(self)
    
    def __deepcopy__(self, memo):
        return BufferDict(self, buf=copy.deepcopy(self.buf, memo))

    def __getstate__(self):
        """ Capture state for pickling. """
        return (_gvar.dumps(collections.OrderedDict(self)), BufferDict.invfcn)

    def __setstate__(self, state):
        """ Restore state when unpickling. """
        if not hasattr(state, 'keys'):
            if isinstance(state, tuple):
                contents, invfcn = state
            else:
                # legacy code
                contents = state 
                invfcn = {}
            tmp = _gvar.BufferDict(_gvar.loads(contents))
            self._odict = tmp._odict 
            self._buf = tmp._buf 
            self._extension = {}
            self.shape = None
            for k in invfcn:
                if k not in self.invfcn:
                    self.invfcn[k] = invfcn[k]
            return
        # even older legacy code
        layout = state['layout']
        buf = state['buf']
        if isinstance(buf, tuple):
            if len(buf) == 2:
                # old format (for legacy)
                buf = _gvar.gvar(*buf)  
            else:
                means, bcovs = buf[:2]
                buf = numpy.array(means, dtype=object)
                for idx, bcov in bcovs:
                    buf[idx] = _gvar.gvar(buf[idx], bcov)
        else:
            buf = numpy.array(buf)
        for k in layout:
            self._odict.__setitem__(
                k,
                BUFFERDICTDATA(
                    slice=layout[k][0],
                    shape=layout[k][1] if layout[k][1] is not None else ()
                    )
                )
        self._buf = buf
        self._extension = {}
        self.shape = None

    def _remove_gvars(self, gvlist):
        tmp = BufferDict(self, buf=numpy.array(self._buf))
        tmp._buf = _gvar.remove_gvars(tmp._buf, gvlist)
        return tmp

    def _distribute_gvars(self, gvlist):
        tmp = BufferDict(self, buf=numpy.array(self._buf))
        tmp._buf = _gvar.distribute_gvars(tmp._buf, gvlist)
        return tmp

    def __reduce_ex__(self, dummy):
        return (BufferDict, (), self.__getstate__())

    def __iadd__(self, g):
        """ self += |BufferDict| (or dictionary) """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        self.flat[:] += g.flat
        return self

    def __isub__(self, g):
        """ self -= |BufferDict| (or dictionary) """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        self.flat[:] -= g.flat
        return self

    def __imul__(self, x):
        """ ``self *= x`` for scalar ``x`` """
        self.flat[:] *= x
        return self

    def __itruediv__(self, x):
        """ ``self /= x`` for scalar ``x`` """
        self.flat[:] /= x
        return self

    def __pos__(self):
        """ ``+self`` """
        return BufferDict(self, buf=+self.flat[:])

    def __neg__(self):
        """ ``-self`` """
        return BufferDict(self, buf=-self.flat[:])

    def __add__(self, g):
        """ :class:`BufferDict` (or a dictionary).

        The two dictionaries need to have compatible layouts: i.e., the
        same keys and array shapes.
        """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        return BufferDict(self, buf=self.flat[:] + g.flat)

    def __radd__(self, g):
        """ Add ``self`` to another :class:`BufferDict` (or a dictionary).

        The two dictionaries need to have compatible layouts: i.e., the
        same keys and array shapes.
        """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        return BufferDict(self, buf=self.flat[:] + g.flat)

    def __sub__(self, g):
        """ Subtract a :class:`BufferDict` (or a dictionary) from ``self``.

        The two dictionaries need to have compatible layouts: i.e., the
        same keys and array shapes.
        """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        return BufferDict(self, buf=self.flat[:] - g.flat)

    def __rsub__(self, g):
        """ Subtract ``self`` from a :class:`BufferDict` (or a dictionary).

        The two dictionaries need to have compatible layouts: i.e., the
        same keys and array shapes.
        """
        try:
            g = BufferDict({k:g[k] for k in self})
        except KeyError:
            raise KeyError('g missing a key')
        return BufferDict(self, buf=g.flat[:] - self.flat)

    def __mul__(self, x):
        """ Multiply ``self``` by scalar ``x``. """
        return BufferDict(self, buf=self.flat[:] * x)

    def __rmul__(self, x):
        """ Multiply ``self`` by scalar ``x``. """
        return BufferDict(self, buf=self.flat[:] * x)

    # truediv and div are the same --- 1st is for python3, 2nd for python2
    def __truediv__(self, x):
        """ Divide ``self`` by scalar ``x``. """
        return BufferDict(self, buf=self.flat[:] / x)

    def __div__(self, x):
        """ Divide ``self`` by scalar ``x``. """
        return BufferDict(self, buf=self.flat[:] / x)

    def add(self,k,v):
        """ Augment buffer with data ``v``, indexed by key ``k``.

        ``v`` is either a scalar or a :mod:`numpy` array (or a list or
        other data type that can be changed into a numpy.array).
        If ``v`` is a :mod:`numpy` array, it can have any shape.

        Same as ``self[k] = v`` except when ``k`` is already used in
        ``self``, in which case a ``ValueError`` is raised.
        """
        if k in self:
            raise ValueError("Key %s already used." % str(k))
        else:
            self[k] = v

    def __getitem__(self,k):
        """ Return piece of buffer corresponding to key ``k``. """
        try:
            d = self._odict.__getitem__(k)
            ans = self._buf[d.slice]
            return ans if d.shape is () else ans.reshape(d.shape)
        except KeyError:
            pass
        if self._ver_l == BufferDict._ver_g:
            try:
                return self._extension[k]
            except KeyError:
                pass
        else:
            self._extension = {}
            self._ver_l = BufferDict._ver_g 
        for f in BufferDict.invfcn:
            altk = f + '(' + str(k) + ')'
            try:
                d = self._odict.__getitem__(altk)
                ans = self._buf[d.slice]
                if d.shape != ():
                    ans = ans.reshape(d.shape)
                ans = BufferDict.invfcn[f](ans)
                self._extension[k] = ans
                return ans
            except KeyError:
                pass
        raise KeyError("undefined key: %s" % str(k))

    def extension_keys(self):
        ans = []
        for k in self:
            try:
                m = re.match(BufferDict.extension_pattern, str(k))
            except TypeError:
                continue
            if m is None:
                continue
            k_fcn, k_stripped = m.groups()
            if k_fcn in BufferDict.invfcn:
                ans.append(k_stripped)
        return ans

    def all_keys(self):
        """ Iterator over all keys and implicit keys.
        
        For example, the following code ::

            b = BufferDict()
            b['log(x)'] = 1.
            for k in b.keys():
                print(k, b[k])
            print()
            for k in b.all_keys():
                print(k, b[k])

        gives the following output::

            log(x) 1

            log(x) 1
            x 2.718281828459045

        Here ``'x'`` is not a key in dictionary ``b`` 
        but ``b['x']`` is defined implicitly from ``b['log(x)']``.
        See :meth:`gvar.BufferDict.add_distribution` for more 
        information.
        """
        # Code from Giacomo Petrillo.
        for k in self:
            yield k
            m = self.extension_pattern.match(k)
            if m and m.group(1) in BufferDict.invfcn:
                yield m.group(2)

    def values(self):
        # needed for python3.5
        return [self[k] for k in self]

    def items(self):
        # needed for python3.5
        return [(k,self[k]) for k in self]

    def __setitem__(self, k, v):
        """ Set piece of buffer corresponding to ``k`` to value ``v``.

        The shape of ``v`` must equal that of ``self[k]`` if key ``k``
        is already in ``self``.
        """
        self._extension = {}
        if k not in self:
            v = numpy.asarray(v, dtype=self._dtype)
            if v.shape==():
                # add single piece of data
                self._odict.__setitem__(
                    k, BUFFERDICTDATA(slice=len(self._buf), shape=())
                    )
                # self._buf = numpy.append(self._buf,v) ####
            else:
                # add array
                n = v.size  # numpy.size(v)    #########
                i = len(self._buf)
                self._odict.__setitem__(
                    k, BUFFERDICTDATA(slice=slice(i,i+n), shape=tuple(v.shape))
                    )
                # self._buf = numpy.append(self._buf, v) #####
            if len(self._buf) == 0:
                self._buf = v.flatten()
            else:
                self._buf = numpy.append(self._buf, v)
        else:
            d = self._odict.__getitem__(k)
            if d.shape is ():
                try:
                    self._buf[d.slice] = v
                except ValueError:
                    raise ValueError(
                        'not a scalar: shape={}'.format(str(numpy.shape(v)))
                        )
                except TypeError:
                    raise TypeError('wrong type: {} not {}'.format(
                        str(type(v)), str(self.dtype)
                        ))
            else:
                v = numpy.asarray(v)
                try:
                    self._buf[d.slice] = v.flat
                except ValueError:
                    raise ValueError('shape mismatch: {} not {}'.format(
                            str(v.shape),str(d.shape)
                            ))

    def __delitem__(self, k):
        if k not in self:
            raise ValueError('key not in BufferDict: ' + str(k))
        size = numpy.size(self[k])
        # fix buffer
        self._buf = numpy.delete(self._buf, self.slice(k), 0)
        # fix slices for keys after k
        keys = list(self.keys())
        idx = keys.index(k)
        for kk in keys[idx + 1:]:
            sl, sh = self._odict.__getitem__(kk)
            if isinstance(sl, slice):
                newsl = slice(sl.start - size, sl.stop - size)
            else:
                newsl = sl - size
            self._odict.__setitem__(
                kk, BUFFERDICTDATA(slice=newsl, shape=sh)
                )
        self._odict.__delitem__(k)

    def __len__(self):
        return len(self._odict)

    def __iter__(self):
        return iter(self._odict)

    def __contains__(self, k):
        return k in self._odict

    def __repr__(self):
        if self._dtype is None:
            return self.__class__.__name__ + '(' + str(self) + ')'
        else:
            return self.__class__.__name__ + '(' + str(self) + ', dtype=' + str(self._dtype) + ')'

    def __str__(self):
        # Code from Giacomo Petrillo.
        out = '{'

        listrepr = [(repr(k), repr(v)) for k, v in self.items()]
        newlinemode = any('\n' in rv for _, rv in listrepr)
        
        for rk, rv in listrepr:
            if not newlinemode:
                out += '{}: {}, '.format(rk, rv)
            elif '\n' in rv:
                rv = rv.replace('\n', '\n    ')
                out += '\n    {}:\n    {},'.format(rk, rv)
            else:
                out += '\n    {}: {},'.format(rk, rv)
                
        if out.endswith(', '):
            out = out[:-2]
        elif newlinemode:
            out += '\n'
        out += '}'
        
        return out

    def _getflat(self):
        self._extension = {}
        return self._buf.flat

    def _setflat(self, buf):
        """ Assigns buffer with buf if same size. """
        self._extension = {}
        self._buf.flat = buf

    flat = property(_getflat, _setflat, doc='Buffer array iterator.')
    def flatten(self):
        """ Copy of buffer array. """
        return numpy.array(self._buf)

    def _getdtype(self):
        return self._buf.dtype

    dtype = property(_getdtype, doc='Data type of buffer array elements.')

    def _getbuf(self):
        self._extension = {}
        return self._buf

    def _setbuf(self, buf):
        """ Replace buffer with ``buf``.

        ``buf`` must be a 1-dimensional :mod:`numpy` array of the same size
        as ``self._buf``.
        """
        self._extension = {}
        if isinstance(buf,numpy.ndarray) and buf.shape == self._buf.shape:
            self._buf = buf
        else:
            raise ValueError(
                "New buffer wrong type or shape ---\n    %s,%s   not   %s,%s"
                % (type(buf), numpy.shape(buf),
                type(self._buf), self._buf.shape))

    buf = property(_getbuf,_setbuf,doc='Buffer array (not a copy).')

    def _getsize(self):
        """ Length of buffer. """
        return len(self._buf)

    size = property(_getsize,doc='Size of buffer array.')

    def slice(self,k):
        """ Return slice/index in ``self.flat`` corresponding to key ``k``."""
        self._extension = {}
        return self._odict.__getitem__(k).slice

    def slice_shape(self,k):
        """ Return tuple ``(slice/index, shape)`` corresponding to key ``k``."""
        self._extension = {}
        return self._odict.__getitem__(k)

    def has_dictkey(self, k):
        """ Returns ``True`` if ``self[k]`` is defined; ``False`` otherwise.

        Note that ``k`` may be a key or it may be related to a
        related key associated with a non-Gaussian distribution
        (e.g., ``'log(k)'``; see :func:`gvar.BufferDict.add_distribution`
        for more information).
        """
        return _gvar.has_dictkey(self, k)

    @staticmethod
    def add_distribution(name, invfcn):
        """ Add new parameter distribution.

        |BufferDict|\s can be used to represent a  restricted
        class of  non-Gaussian distributions. For example, the code ::

            import gvar as gv
            gv.BufferDict.add_distribution('log', gv.exp)

        enables the use of log-normal distributions for parameters. So
        defining, for example, ::

            b = gv.BufferDict()
            b['log(a)'] = gv.gvar('1(1)')

        means that ``b['a']`` has a value (equal to ``exp(b['log(a)']``)
        even though ``'a'`` is not a key in the dictionary.

        The distributions available by default correspond to::

            gv.BufferDict.add_distribution('log', gv.exp)
            gv.BufferDict.add_distribution('sqrt', gv.square)
            gv.BufferDict.add_distribution('erfinv', gv.erf)

        Args:
            name (str): Distributions' function name. A ``ValueError`` is 
                raised if ``name`` is already being used for 
                a distribution; the error can be avoided by deleting
                the old definition first using 
                :meth:`BufferDict.del_distribution`.
            invfcn (callable): Inverse of the transformation function.
        """
        if name in BufferDict.invfcn:
            raise ValueError('distribution {} already defined'.format(name))
        BufferDict.invfcn[name] = invfcn

    @staticmethod
    def del_distribution(name):
        """ Delete |BufferDict| distribution ``name``. 
        
        Raises a ``ValueError`` if ``name`` is not the name of 
        an existing distribution.
        """
        if name in BufferDict.invfcn:
            del BufferDict.invfcn[name]
        else:
            raise ValueError('{} is not a distribution'.format(name))
        BufferDict._ver_g += 1

    @staticmethod
    def has_distribution(name):
        """ ``True`` if ``name`` has been defined as a distribution; ``False`` otherwise. """
        return name in BufferDict.invfcn
    
    @staticmethod
    def uniform(fname, umin, umax, shape=()):
        """ Create uniform distribution on interval ``[umin, umax]``.

        The code ::

            import gvar as gv
            b = gv.BufferDict()
            b['f(w)'] = gv.BufferDict.uniform('f', 2., 3.)
        
        adds a distribution function ``f(w)`` designed so that ``b['w']``
        corresponds to a uniform distribution on the interval ``[2., 3.]``
        (see :meth:`gvar.BufferDict.add_distribution` for more about 
        distributions).

        Args:
            fname (str): Name of function used in the :class:`BufferDict` key. 
                Note that names can be reused provided they correspond to the 
                same interval as in previous calls.
            umin (float): Minimum value of the uniform distribution.
            umax (float): Maximum value of the uniform distribution.
            shape (tuple): Shape of array of uniform variables. Default is ``()``.
        
        Returns:
            :class:`gvar.GVar` object corresponding to a uniform distribution.
        """
        if not isinstance(fname, str):
            raise ValueError('fname must be a string')
        if fname in BufferDict.invfcn:
            invfcn = BufferDict.invfcn[fname]
            if not isinstance(invfcn, _BDict_UDistribution):
                raise ValueError("distribution {} already defined".format(fname))
            elif sorted((invfcn.umin, invfcn.umax)) != sorted((umin, umax)):
                raise ValueError("distribution {} already defined".format(fname))
        else:
            BufferDict.add_distribution(fname, _BDict_UDistribution(umin, umax))
        if shape == ():
            return _gvar.gvar(0, 1)
        else:
            ans = _gvar.gvar(int(numpy.prod(shape)) * [(0, 1.)]) 
            ans.shape = shape 
            return ans

class _BDict_UDistribution(object):
    def __init__(self, umin, umax):
        self.umin = umin 
        self.umax = umax
        self.umax_umin_2 = (umax - umin) / 2.
        self.root2 = numpy.sqrt(2)
    def __call__(self, x):
        return self.umin + (_gvar.erf(x / self.root2) + 1) * self.umax_umin_2

BufferDict._ver_g = 0    # version of distribution collection (for cache synch)

def asbufferdict(g, dtype=None):
    """ Convert ``g`` to a BufferDict, keeping only ``g[k]`` for ``k in keylist``.

    ``asbufferdict(g)`` will return ``g`` if it is already a
    :class:`gvar.BufferDict`; otherwise it will convert the dictionary-like
    object into a :class:`gvar.BufferDict`. The data can also be
    specified: e.g., ``asbufferdict(g, dtype=int)``.
    """
    if isinstance(g, BufferDict) and dtype is None:
        return g
    kargs = {} if dtype is None else dict(dtype=dtype)
    return BufferDict(g, **kargs)

def get_dictkeys(bdict, klist):
    """ Same as ``[dictkey(bdict, k) for k in klist]``. """
    ans = []
    for k in klist:
        if k not in bdict:
            for f in BufferDict.invfcn:
                newk = f + '(' + str(k) + ')'
                if newk in bdict:
                    break
            else:
                raise KeyError('bad key: ' + str(k))
            ans.append(newk)
        else:
            ans.append(k)
    return ans

def dictkey(bdict, k):
    """ Find key in ``bdict`` corresponding to ``k``.

    Could be ``k`` itself or one of the standard extensions,
    such as ``log(k)`` or ``sqrt(k)``.
    """
    if k not in bdict:
        for f in BufferDict.invfcn:
            newk = f + '(' + str(k) + ')'
            if newk in bdict:
                return newk
        raise KeyError('key not used: ' + str(k))
    else:
        return k

def has_dictkey(b, k):
    """ Returns ``True`` if ``b[k]`` is defined; ``False`` otherwise.

    Note that ``k`` may be a key or it may be related to a
    related key associated with a non-Gaussian distribution
    (e.g., ``'log(k)'``; see :func:`gvar.MultiFitter.add_distribution`).
    """
    if k in b:
        return True
    else:
        for f in BufferDict.invfcn:
            newk = f + '(' + str(k) + ')'
            if newk in b:
                return True
        return False

def _stripkey(k):
    """ Return (stripped key, fcn) where fcn is exp or square or ...

    Strip off any ``"log"`` or ``"sqrt"`` or ... prefix.
    """
    if not isinstance(k, str):
        return k, None
    m = re.match(BufferDict.extension_pattern, k)
    if m is None:
        return k, None
    k_fcn, k_stripped = m.groups()
    if k_fcn not in BufferDict.invfcn:
        return k, None
    return k_stripped, BufferDict.invfcn[k_fcn]


def nonredundant_keys(keys):
    """ Return list containing only nonredundant keys in list ``keys``. """
    discards = set()
    for k in keys:
        if isinstance(k, str):
            m = re.match(BufferDict.extension_pattern, k)
            if m is not None:
                discards.add(m.groups()[1])
    ans = []
    for k in keys:
        if not isinstance(k, str) or k not in discards:
            ans.append(k)
    return ans

def add_parameter_parentheses(p):
    """ Return dictionary with proper keys for parameter distributions (legacy code).

    This utility function helps fix legacy code that uses
    parameter keys like ``logp`` or ``sqrtp`` instead of
    ``log(p)`` or ``sqrt(p)``, as now required. This method creates a
    copy of  dictionary ``p'' but with keys like ``logp`` or ``sqrtp``
    replaced by ``log(p)`` or ``sqrt(p)``. So setting ::

        p = add_parameter_parentheses(p)

    fixes the keys in ``p`` for log-normal and sqrt-normal parameters.
    """
    newp = BufferDict()
    for k in p:
        if isinstance(k, str):
            if k[:3] == 'log' and _stripkey(k)[1] is None:
                newk = 'log(' + k[3:] + ')'
            elif k[:4] == 'sqrt' and _stripkey(k)[1] is None:
                newk = 'sqrt(' + k[4:] + ')'
            else:
                newk = k
            newp[newk] = p[k]
    return newp

def trim_redundant_keys(p):
    """ Remove redundant keys from dictionary ``p``.

    A key ``'c'`` is redundant if either of ``'log(c)'``
    or ``'sqrt(c)'`` is also a key. (There are additional redundancies
    if :meth:`gvar.add_parameter_distribution` has been used to add
    extra distributions.) This function creates a copy of ``p`` but with
    the redundant keys removed.
    """
    return BufferDict(p, keys=nonredundant_keys(p.keys()))



""" Pade approximants for GVars.  """
# Created by G. Peter Lepage on 2009-12-14.
# Copyright (c) 2009-2020 G. Peter Lepage.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version (see <http://www.gnu.org/licenses/>).
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import sys
import numpy 
import gvar as _gvar 
# import scipy.linalg 
from scipy.interpolate import pade as _scipy_pade

class Pade(object):
    """ Pade approximant to ``sum_i f[i] x**i`` for ``GVar``\s.

    The ``order=(m,n)`` Pade approximant to a series given by
    ``sum_i f[i] * x**i`` is the ratio of  polynomials of order ``m``
    (numerator) and ``n`` (denominator) whose  Taylor expansion agrees
    with that of the original series up to order ``m+n``.

    A :class:`Pade` object ``pade`` creates :class:`gvar.powerseries.PowerSeries`
    objects for the numerator (``pade.num``) and the denominator (``pade.den``)
    of the Pade approximant corresponding to the input series ``f[i]``. The 
    approximant can be evaluated for arbitrary ``x`` using ``pade(x)``. The
    coefficients used in the numerator and denominator are given by ``pade.num.c``
    and ``pade.den.c``, respectively.

    Elements in the series ``f[i]`` may be numbers or :class:`gvar.GVar`\s.
    When the latter appear, the code uses an SVD algorithm (see :func:`pade_svd`) 
    to deal with the imprecision in the input data. It automatically reduces
    the order of the approximant if the extraction of Pade coefficients
    is too unstable given the noise in the input data. The actual order used 
    in the approximant is given by ``pade.order``.

    Examples:
        Typical usage is::

            >>> import gvar as gv
            >>> c = gv.gvar(['1(0)', '1.0(1)', '0.500(10)', '0.1667(100)','.04167(100)'])
            >>> pade = gv.pade.Pade(c, order=(2,2))
            >>> print(pade.num.c, pade.den.c)               # num and den coefficients
            [1(0) 0.50(14) 0.08(11)] [1(0) -0.50(14) 0.083(55)]
            >>> print(pade(1))                              # evaluate pade at x=1
            2.714(99)
            >>> print(pade.num(1), pade.den(1), pade.order) # evaluate num and den at x=1
            1.58(21) 0.583(82) (2, 2) 

        When errors are larger (see ``c[0]``), the order may be automatically 
        reduced (here to ``(1,2)``)::

            >>> c = gv.gvar(['1.0(1)', '1.0(1)', '0.500(10)', '0.1667(100)','.04167(100)'])
            >>> pade = gv.pade.Pade(c, order=(2,2))
            >>> print(pade.num.c, pade.den.c)               # num and den coefficients
            [1.00(10) 0.33(29)] [1(0) -0.67(17) 0.17(11)]
            >>> print(pade(1))                              # evaluate pade at x=1
            2.67(20)
            >>> print(pade.num(1), pade.den(1), pade.order) # evaluate num and den at x=1
            1.33(27) 0.500(69) (1, 2)
 
    Args:
        f: Array ``f[i]`` of power series coefficients for ``i=0...n+m``.
        order (tuple): ``order=(m,n)`` specifies that the numerator and 
            denominator of the Pade approximant have maximum order 
            ``m`` and ``n``, respectively.
        rtol (float or str): Relative uncertainty in the coefficients, for
            use by the SVD algorithm. If ``rtol`` is a string, it determines 
            how the relative tolerance is determined from the relative
            uncertainties in the ``f[i]``. Set ``rtol`` equal to:
            ``'gavg'`` for the geometric mean (default); ``'avg'`` for
            the average; ``'min'`` for the minimum; or ``'max'`` for
            the maximum. Otherwise a number can be specified, in which case
            the uncertainties in ``f[i]`` are ignored. The SVD analysis
            is skipped if ``rtol=None``.
    """
    def __init__(self, f, order, rtol='gavg'):
        # check inputs
        m, n = order
        assert m >= 0 and n >= 0
        f = f[:n + m + 1]
        if len(f) < (m + n + 1):
            raise ValueError(
                'not enough f[i]s -- need {} have {}'.format(n + m + 1, len(f))
                )
        if not numpy.any([isinstance(fi, _gvar.GVar) for fi in f]):
            if rtol is None:
                p, q = Pade._scipy_pade(f, n)
            else:
                self.rtol = 1e-14
                p, q = pade_svd(f, m, n, rtol=self.rtol)
            self.num = _gvar.powerseries.PowerSeries(p)
            self.den = _gvar.powerseries.PowerSeries(q)
            self.order = [len(p) - 1, len(q) - 1]
            return
        else:
            c = numpy.array(f)

        # compute tolerance if not specified
        if rtol in ['avg', 'min', 'max', 'gavg']:
            means = numpy.fabs(_gvar.mean(c))
            sdevs = _gvar.sdev(c)
            idx = means > 0.0
            if numpy.any(idx) and numpy.all(sdevs[idx] > 0):
                ratio = sdevs[idx] / means[idx]
                if rtol == 'gavg':
                    # geometric mean
                    rtol = numpy.exp(
                        numpy.average(numpy.log(ratio))
                        )
                elif rtol == 'avg':
                    rtol = numpy.average(ratio)
                elif rtol == 'min':
                    rtol = numpy.min(ratio)
                else:
                    rtol = numpy.max(ratio)
            else:
                rtol = 1e-14
        elif rtol is not None:
            rtol = numpy.fabs(rtol)
        self.rtol = rtol

        # find Pade coefficients
        if self.rtol is None:
            p, q = Pade._scipy_pade(_gvar.mean(c), n)
        else:
            p, q = pade_svd(_gvar.mean(c), m, n, rtol=self.rtol)
        m = len(p) - 1
        n = len(q) - 1

        # add uncertainties
        p = p * _gvar.gvar(len(p) * ['1(0)'])
        q = q[1:] * _gvar.gvar(len(q[1:]) * ['1(0)'])
        num = _gvar.powerseries.PowerSeries(p, order=m + n)
        den = _gvar.powerseries.PowerSeries([1] + list(q), order=m + n)
        pq = numpy.concatenate((p,q))
        cc = (num / den).c
        M = numpy.empty((len(pq), len(pq)), float)
        for i in range(len(pq)):
            for j in range(len(pq)):
                M[i, j] = cc[i].deriv(pq[j])
        pq = pq + _gvar.linalg.solve(M, (c - _gvar.mean(c))[:len(pq)])
        self.num = _gvar.powerseries.PowerSeries(pq[:m + 1])
        self.den = _gvar.powerseries.PowerSeries([_gvar.gvar(1,0)] + list(pq[m + 1:]))
        self.order = (m, n)

    @staticmethod
    def _scipy_pade(c, n):
        p, q = _scipy_pade(c, n)
        return numpy.array(p.c[::-1]), numpy.array(q.c[::-1])

    def __call__(self, x):
        return self.num(x) / self.den(x)

def pade_gvar(f, m, n, rtol='gavg'):  
    """ ``(m,n)`` Pade approximant to ``sum_i f[i] x**i`` for ``GVar``\s.

    The ``(m,n)`` Pade approximant to a series given by
    ``sum_i f[i] * x**i`` is the ratio of  polynomials of order ``m``
    (numerator) and ``n`` (denominator) whose  Taylor expansion agrees
    with that of the original series up to order ``m+n``.

    This code uses an SVD algorithm (see :func:`pade_svd`) to deal with
    imprecision in the input data. It automatically reduces
    the order of the approximant if the extraction of Pade coefficients
    is too unstable given noise in the input data.

    Args:
        f: Array ``f[i]`` of power series coefficients for ``i=0...n+m``.
        m: Maximum order of polynomial in numerator of Pade
            approximant (``m>=0``).
        n: Maximum order of polynomial in denominator of Pade
            approximant (``m>=0``).
        rtol (float or str): If ``rtol`` is a string, it determines how the
            relative tolerance is determined from the relative
            uncertainties in the ``f[i]``. Set ``rtol`` equal to:
            ``'gavg'`` for the geometric mean (default); ``'avg'`` for
            the average; ``'min'`` for the minimum; or ``'max'`` for
            the maximum. Otherwise a number can be specified, in which case
            the uncertainties in ``f[i]`` are ignored. The SVD analysis is 
            skipped if ``rtol=None``.
    Returns:
        Tuple of power series coefficients ``(p, q)`` such that
        ``sum_i p[i] x**i`` is the numerator of the approximant,
        and ``sum_i q[i] x**i`` is the denominator. ``q[0]`` is
        normalized to 1.
    """
    pade = Pade(f, order=(m,n), rtol=rtol)
    return (pade.num.c, pade.den.c)

def pade_svd(f, m, n, rtol=1e-14):
    """ ``(m,n)`` Pade approximant to ``sum_i f[i] x**i``.

    The ``(m,n)`` Pade approximant to a series given by
    ``sum_i f[i] * x**i`` is the ratio of  polynomials of order ``m``
    (numerator) and ``n`` (denominator) whose  Taylor expansion agrees
    with that of the original series up to order ``m+n``.

    This code is adapted from P. Gonnet,  S. Guttel, L. N. Trefethen, SIAM
    Review Vol 55, No. 1, 101 (2013). It uses an SVD algorithm to deal with
    imprecision in the input data,  here specified by the relative tolerance
    ``rtol`` for the  input coefficients ``f[i]``. It automatically reduces
    the order of the approximant if the extraction of Pade coefficients
    is too unstable given tolerance ``rtol``.

    Args:
        f: Array ``f[i]`` of power series coefficients for ``i=0...n+m``.
        m: Maximum order of polynomial in numerator of Pade
            approximant (``m>=0``).
        n: Maximum order of polynomial in denominator of Pade
            approximant (``m>=0``).
        rtol: Relative accuracy of input coefficients. (Default is 1e-14.)
    
    Returns:
        Tuple of power series coefficients ``(p, q)`` such that
        ``sum_i p[i] x**i`` is the numerator of the approximant,
        and ``sum_i q[i] x**i`` is the denominator. ``q[0]`` is
        normalized to 1.
    """
    import scipy.linalg as linalg
    # linalg = scipy.linalg
    mn_save = m,n
    c = numpy.array(f[:n + m + 1], float)
    if len(f) < (m + n + 1):
        raise ValueError(
            'not enough f[i]s -- need {} have {}'.format(n + m + 1, len(f))
            )
    # if USE_SCIPY_PADE:
    #     p, q = scipy_pade(c, n)
    #     return numpy.array(p.c[::-1]), numpy.array(q.c[::-1])
    ts = rtol * linalg.norm(c)
    if linalg.norm(c[:m + 1]) <= rtol * linalg.norm(c):
        # return power series through order m
        a = numpy.array(c[:1])
        b = numpy.array([1.])
    else:
        row = numpy.zeros(n+1)
        row[0] = c[0]
        col = c
        while True:
            if n == 0:
                # return the power series through order m
                a = c[:m + 1]
                b = numpy.array([1.])
                return a, b
            Z = linalg.toeplitz(col[:m + n + 1], row[:n + 1])
            C = Z[m + 1:, :]
            rho = numpy.sum(linalg.svdvals(C) > ts)
            if rho == n:
                break
            m -= n - rho
            n = rho
            if m < 0:
                m = 0
        if n > 0:
            # use svd to get solution b, but only to normalize C
            # then use QR decomposition to get final b
            U, S, V = linalg.svd(C, full_matrices=True)
            b = V.transpose()[:, -1]
            D = numpy.diag(numpy.abs(b) + numpy.sqrt(sys.float_info.epsilon))
            Q,R = linalg.qr(C.dot(D).transpose())
            b = D.dot(Q)[:,-1]
            b = b / linalg.norm(b)
            a = Z[:m + 1, :n + 1].dot(b)
            lam = numpy.where(abs(b) > rtol)[0][0]
            b = b[lam:]
            a = a[lam:]
            b = b[:numpy.where(abs(b) > rtol)[0][-1] + 1]
        idx = abs(a) > ts
        if not numpy.any(idx):
            a = a[:1]
        else:
            a = a[:numpy.where(idx)[0][-1] + 1]
        a = a / b[0]
        b = b / b[0]
    # N.B.: an approximant for non-zero rtol is the
    # same as the reduced-order approximant evaluated with
    # zero rtol; any approximant returned by the algorithm should
    # be an exact approximant to the input. Thus rtol determines the
    # order of the final approximant, but does not affect the values
    # of the approximant's coefficients.
    mfinal = len(a) - 1
    nfinal = len(b) - 1
    return (a,b) if (mfinal,nfinal) == mn_save else pade_svd(f, mfinal, nfinal)



from setuptools import setup, Extension
import numpy

ext_args = dict(
    include_dirs=[numpy.get_include()],
    library_dirs=[],
    runtime_library_dirs=[],
    extra_link_args=[]
    )

ext_modules = [
    Extension(name="gvar._gvarcore", sources=["src/gvar/_gvarcore.pyx"], **ext_args),
    Extension(name= "gvar._svec_smat", sources=["src/gvar/_svec_smat.pyx"], **ext_args),
    Extension(name="gvar._utilities", sources=["src/gvar/_utilities.pyx"], **ext_args),
    Extension(name="gvar.dataset", sources=["src/gvar/dataset.pyx"], **ext_args),
    Extension(name="gvar._bufferdict", sources=["src/gvar/_bufferdict.pyx"], **ext_args),
    ]

setup(ext_modules=ext_modules)
